       PROGRAM RFLUX
C===============================================================================
C-- After NUCLEO is run, this program does the calculations necessary
C-- for the network flux plot.  It reads NUCLEO.DAT to get the isotope
C-- list and the reaction list.  It reads TOTFLUX.OUT to get the integrated
C-- flux for each reaction.  It calculates the net flow between every pair
C-- of nuclei, and assigns an index based on its magnitude.  If:
C--         flow > maxflow/10,   index = 3
C--         flow > maxflow/100,  index = 2
C--         flow > maxflow/1000, index = 1
C--         otherwise, index = 0
C-- Output goes to RFLUX.OUT, where a list of 
C-- isotopes is written, followed by a list of all the pairs of nuclei
C-- with an index greater than 0.  
C===============================================================================
       integer ret
       ret=0
       call fluxSub(ret)
       end

c=======================================================================
       subroutine fluxSub(ret)


       INTEGER ret
       DOUBLE PRECISION TOTFLUX(50000),MAXFLUX
       DOUBLE PRECISION NETFLUX(50000)
       INTEGER NIS(3000),ZIS(3000),AIS(3000)
       INTEGER INN(50000),INZ(50000),OUTN(50000),OUTZ(50000)
       INTEGER GOTSPEC,fluxindex,DONE(50000)
       INTEGER NET,NRE,NSL,I,K,KK,J,ja,jb
       INTEGER NA,NB,NC,ND
       CHARACTER*(5) ISO(3000),INSPEC(50000),OUTSPEC(50000)
       CHARACTER*(5) NEWISO(3000)
       CHARACTER*(5) SPEC1,SPEC2
       character*30 networkfile,fluxfile
       character*80 ABCD
       character*5 efgh
       character*11 zzz
       character*1 response

       networkfile='nucleo.dat'
       fluxfile='totflux.out'
       
c       write(6,9333) networkfile
c       read(5,9444) response
c       if ((index(response,'y').ne.0).or.
c     *    (index(response,'Y').ne.0)) then
c          write(6,*)' Enter new filename:'
c	  read(5,9444) networkfile
c       endif
c9333   format(/,10x,'Change network input file? [Y/N]',/,10x,      
c     *         '(Default: ',a,' )')
       
c       write(6,9555) fluxfile
c       read(5,9444) response
c       if ((index(response,'y').ne.0).or.
c     *    (index(response,'Y').ne.0)) then
c          write(6,*)' Enter new filename:'
c	  read(5,9444) fluxfile
c       endif
c9555   format(/,10x,'Change flux input file? [Y/N]',/,10x,      
c     *         '(Default: ',a,' )')
c9444   format(a)
              	             
       OPEN(8,FILE=networkfile,STATUS='OLD')  ! INPUT FILE
       OPEN(10,FILE=fluxfile,STATUS='OLD')! INPUT FILE
       OPEN(21,FILE='rflux.out',STATUS='unknown')   ! OUTPUT FILE
c       open(22,file='rflux_test.out',STATUS='unknown')

C -----------------------------------------------------------
C -- GET THE ISOTOPE LIST                                   
C -----------------------------------------------------------
       READ(8,8101) NET, ABCD
 8101  FORMAT(I6,A80)
       READ(8,8100) ABCD
 8100  FORMAT(A80)

C --OUTPUT NUMBER OF ISOTOPES
       WRITE(21,7000) NET
 7000  FORMAT(1X,I5)
c       WRITE(6,*)'testing'

C -- Read isotopes, calculate N, get string for output to the graph.
       DO 40 I=1,NET
         READ(8,9025) KK,ISO(I),AIS(I),ZIS(I)
 9025    FORMAT(I4,1X,A5,1X,i3,1x,i3)
         NIS(I)=AIS(I)-ZIS(I)
         NEWISO(I)=ISO(I)
         IF (ZIS(I).EQ.0) NEWISO(I)(1:2) = 'n '
         IF (NEWISO(I)(2:2).NE.' ') THEN
           NEWISO(I)(2:2)=CHAR(ICHAR(NEWISO(I)(2:2))+32)
         ENDIF
         WRITE(21,9030) I,NEWISO(I),AIS(I),ZIS(I),NIS(I)
 9030    FORMAT(I4,',',1X,A5,',',1X,i4,',',1x,i4,',',1x,i4)
   40  CONTINUE

C -----------------------------------------------------------
C -- SKIP OVER BINDING ENERGIES AND ABUNDANCES
C -----------------------------------------------------------
       READ(8,8102) NSL, ABCD
 8102  format(i6,a80)
       DO 50 I=1,NSL
         READ(8,8100) ABCD
   50  CONTINUE

C -----------------------------------------------------------
C -- GET REACTION LIST
C -----------------------------------------------------------
       READ(8,9033) NRE, ABCD
c       print*,NRE
 9033  FORMAT(I6,A80)

C --OUTPUT NUMBER OF REACTIONS
       WRITE(21,7000) NRE

       DO 60 I=1,NRE
         READ(8,9035) KK,NA,INSPEC(I),NB,SPEC1,NC,SPEC2,ND,
     *    OUTSPEC(I),zzz
 9035    FORMAT(I5,3X,I2,1X,A5,2X,I2,1X,A5,1X,I2,1X,A5,2X,
     *   I2,1X,A5,3X,a11)
   60  CONTINUE

C -----------------------------------------------------------
C -- READ INTEGRATED FLUXES FROM TOTFLUX.OUT
C -----------------------------------------------------------
       DO 62 I=1,NRE
          READ(10,9042) efgh,TOTFLUX(I)
c          write(6,*) totflux(i)
   62  continue
 9042  FORMAT(a5,2x,D21.15)

C -----------------------------------------------------------
C -- FIND MAXIMUM REACTION FLUX
C -----------------------------------------------------------
       MAXFLUX = 0.d0
       DO 27 J = 1,NRE
         IF (TOTFLUX(J).GT.MAXFLUX) THEN 
           MAXFLUX=TOTFLUX(J)
         ENDIF
   27  CONTINUE

       WRITE(6,*) 'MAXTOTFLUX=', MAXFLUX

C -----------------------------------------------------------
C -- MATCH REACTIONS WITH ISOTOPES
C -----------------------------------------------------------
       DO 70 I=1,NRE

         GOTSPEC = 0

         DO 80 K=1,NET
            IF (INSPEC(I).EQ.ISO(K)) THEN
               INN(I)=NIS(K)
               INZ(I)=ZIS(K)
               GOTSPEC=GOTSPEC+1
            ENDIF
            IF (OUTSPEC(I).EQ.ISO(K)) THEN
               OUTN(I)=NIS(K)
               OUTZ(I)=ZIS(K)
               GOTSPEC=GOTSPEC+1
            ENDIF
   80    CONTINUE
C --Check if both isotopes are found.
         IF (GOTSPEC.NE.2) THEN 
           WRITE(6,*) 'ERROR:ISOTOPE NOT FOUND'
           WRITE(6,*) 'CHECK EQUATION:',I
         ENDIF

   70  CONTINUE

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C Here begins the second part of the program: output to RFLUX.OUT
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C-- Initialize Array DONE. Tells us if we have already dealt with reac.
       DO 85 K=1,NRE
          DONE(K)=0
  85   continue
       ja=0
       jb=0

       do 90 k=1,nre
          ja=k
          if(done(k).eq.0) then
             do 100 i=k+1,nre 
                if((inspec(k).eq.inspec(i)).and.
     *           (outspec(k).eq.outspec(i))) then
c                   write(22,*) k,totflux(k)
c                   write(22,*) i,totflux(i)
c                   write(22,*) '11111111111111'
                   if(dabs(totflux(i)).ge.dabs(totflux(k))) then
                      ja=i
                   endif
                   totflux(k)=totflux(k)+totflux(i)
                   done(i)=1
                 elseif((inspec(k).eq.outspec(i)).and.
     *           (outspec(k).eq.inspec(i))) then
c                   write(22,*) k,totflux(k)
c                   write(22,*) i,totflux(i)
c                   write(22,*) '22222222222222'
                   if(dabs(totflux(i)).ge.dabs(totflux(k))) then
                      jb=i
                   endif
                   totflux(k)=totflux(k)-totflux(i)
                   done(i)=1
                 endif
 100         continue
             if(totflux(k).ge.0.d0) then             
                netflux(ja)=totflux(k)
c                write(22,*) ja,netflux(ja)
c                write(22,*) '3333333333333333'
                ja=0
             else
                netflux(k)=0.d0
                netflux(jb)=-1.d0*totflux(k)
c                write(22,*) jb,netflux(jb)
c                write(22,*) '444444444444444'
                jb=0
             endif
          endif      
 90    continue

C --FIND MAXIMUM REACTION FLUX
       MAXFLUX = 0.d0
       DO 110 J = 1,NRE
         IF ((NETFLUX(J).GT.MAXFLUX).and.((INN(J).ne.13).or.
     *   (INZ(J).ne.13).or.(OUTN(J).ne.13).or.(OUTZ(J).ne.13))) 
     *    THEN 
           MAXFLUX=NETFLUX(J)
         ENDIF
  110  CONTINUE
       WRITE(*,*) 'MAXNETFLUX=',MAXFLUX

C --OUTPUT REACTION N,Z
       fluxindex=1
       DO 120 K=1,NRE
c           IF (NETFLUX(K).GE.(MAXFLUX/10.d0)) THEN
c             FLUXINDEX = 4
c           else IF (NETFLUX(K).GE.(MAXFLUX/1000.d0)) THEN
c             FLUXINDEX = 3
c           else IF (NETFLUX(K).GE.(MAXFLUX/100000000000000.d0)) THEN
c             FLUXINDEX = 2
c           ELSE
c             FLUXINDEX = 1
c           ENDIF
           WRITE(21,8061)
     *     K,INN(K),INZ(K),OUTN(K),OUTZ(K),FLUXINDEX,NETFLUX(K)
c          IF(FLUXINDEX.NE.0) WRITE(21,8061)
c    *     K,INN(K),INZ(K),OUTN(K),OUTZ(K),FLUXINDEX,NETFLUX(K)

120   CONTINUE

 8061  FORMAT(i5,',',1X,I3,',',1X,I3,',',1X,I3,',',1X,I3,',',1X,I2,
     *   ',',1X,D14.7)

c      pause
       CLOSE(8)
       CLOSE(10)
       CLOSE(21)

       RETURN
      END
